;;; Mudsync --- Live hackable MUD
;;; Copyright © 2016 Christopher Allan Webber <cwebber@dustycloud.org>
;;;
;;; This file is part of Mudsync.
;;;
;;; Mudsync is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Mudsync is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Mudsync.  If not, see <http://www.gnu.org/licenses/>.

;;; Game actor
;;; ==========

(define-module (mudsync gameobj)
  #:use-module (mudsync command)
  #:use-module (8sync systems actors)
  #:use-module (8sync agenda)
  #:use-module (srfi srfi-1)
  #:use-module (ice-9 match)
  #:use-module (oop goops)
  #:export (<gameobj>
            gameobj-simple-name-f

            gameobj-loc
            gameobj-gm
            gameobj-name
            gameobj-name-f

            gameobj-actions))

;;; Gameobj
;;; =======


;;; Actions supported by all gameobj
(define gameobj-actions
  (build-actions
   (get-commands (wrap-apply gameobj-get-commands))
   (get-container-commands (wrap-apply gameobj-get-container-commands))
   (get-occupants (wrap-apply gameobj-get-occupants))
   (add-occupant! (wrap-apply gameobj-add-occupant!))
   (remove-occupant! (wrap-apply gameobj-remove-occupant!))
   (set-loc! (wrap-apply gameobj-set-loc!))
   (get-name (wrap-apply gameobj-get-name))
   (get-desc (wrap-apply gameobj-get-desc))
   ))

;;; *all* game components that talk to players should somehow
;;; derive from this class.
;;; And all of them need a GM!

(define-class <gameobj> (<actor>)
  ;; location id
  (loc #:init-value #f
       #:accessor gameobj-loc)
  
  ;; Uses a hash table like a set (values ignored)
  (occupants #:init-thunk make-hash-table
             #:accessor gameobj-occupants)

  ;; game master id
  (gm #:init-keyword #:gm
      #:getter gameobj-gm)
  ;; a name to be known by
  (name #:init-keyword #:name
        #:accessor gameobj-name)

  (desc #:init-value ""
        #:init-keyword #:desc)

  ;; how to print our name
  (name-f #:init-keyword #:name-f
          #:getter gameobj-name-f
          #:init-value (wrap gameobj-simple-name-f))

  ;; Name aliases
  (aliases #:init-keyword #:aliases
           #:init-value '())

  ;; Commands we can handle
  (commands #:init-value '())

  ;; Commands we can handle by being something's container
  (container-commands #:init-value '())
  (message-handler
   #:init-value
   (simple-dispatcher gameobj-actions)))


;;; gameobj message handlers
;;; ========================

;; Kind of a useful utility, maybe?
(define (simple-slot-getter slot)
  (lambda (actor message)
    (reply-message actor message
                   #:val (slot-ref actor slot))))

(define (val-or-run val-or-proc)
  "Evaluate if a procedure, or just return otherwise"
  (if (procedure? val-or-proc)
      (val-or-proc)
      val-or-proc))

(define (filter-commands commands verb)
  (filter
   (lambda (cmd)
     (equal? (command-verbs cmd)
             verb))
   commands))

(define-mhandler (gameobj-get-commands actor message verb)
  (define filtered-commands
    (filter-commands (val-or-run (slot-ref actor 'commands))
                     verb))
  (<-reply actor message #:commands filtered-commands))

(define-mhandler (gameobj-get-container-commands actor message verb)
  (define filtered-commands
    (filter-commands (val-or-run (slot-ref actor 'container-commands))
                     verb))
  (<-reply actor message #:commands filtered-commands))

(define-mhandler (gameobj-add-occupant! actor message who)
  (hash-set! (slot-ref actor 'occupants)
             who #t))

(define-mhandler (gameobj-remove-occupant! actor message who)
  (hash-remove! (slot-ref actor 'occupants) who))

(define-mhandler (gameobj-get-occupants actor message)
  (define occupants
    (hash-map->list (lambda (key val) key)
                    (gameobj-occupants actor)))

  (<-reply actor message
           #:occupants occupants))

;; @@: Should it really be #:id ?  Maybe #:loc-id or #:loc?
(define-mhandler (gameobj-set-loc! actor message loc)
  (define old-loc (gameobj-loc actor))
  (format #t "DEBUG: Location set to ~s for ~s\n"
          loc (actor-id-actor actor))

  (set! (gameobj-loc actor) loc)
  ;; Change registation of where we currently are
  (if loc
      (<-wait actor loc 'add-occupant! #:who (actor-id actor)))
  (if old-loc
      (<-wait actor old-loc 'remove-occupant! #:who (actor-id actor))))

(define gameobj-get-name (simple-slot-getter 'name))
(define gameobj-get-desc (simple-slot-getter 'desc))

(define (gameobj-simple-name-f gameobj)
  "Simplest version: return ourselves for our name."
  (gameobj-name gameobj))


